# Copyright Red Hat
#
# snapm/manager/plugins/__init__.py - Snapshot Manager plugins
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
"""
Snapshot manager plugin helpers.
"""
import os
from os.path import sep as path_sep, ismount

from snapm import SNAPM_VALID_NAME_CHARS

_MOUNT_SEPARATOR = "-"
_ESCAPED_MOUNT_SEPARATOR = "--"

_PROC_MOUNTS = "/proc/self/mounts"

DEV_PREFIX = "/dev"
DEV_MAPPER_PREFIX = "/dev/mapper/"

# dmsetup options
DMSETUP_CMD = "dmsetup"
DMSETUP_SPLITNAME = "splitname"
DMSETUP_INFO = "info"
DMSETUP_LVM_SUBSYS = "LVM"
DMSETUP_NO_HEADINGS = "--noheadings"
DMSETUP_COLUMNS = "--columns"
DMSETUP_FIELDS_VG_LV = "-ovg_name,lv_name"
DMSETUP_FIELDS_UUID = "-ouuid"
DMSETUP_REPORT_SEP = ":"

# Fields: origin_name-snapset_snapset-name_timestamp_mount-point
SNAPSHOT_NAME_FORMAT = "%s-snapset_%s_%d_%s"


def _escape_bad_chars(path):
    """
    Encode illegal characters in mount point path.

    :param path: The path to escape.
    :returns: The escaped path.
    """
    escaped = ""
    for char in path:
        if char != path_sep and char not in SNAPM_VALID_NAME_CHARS:
            escaped = escaped + "." + char.encode("utf8").hex()
        else:
            if char == ".":
                escaped = escaped + ".."
            else:
                escaped = escaped + char
    return escaped


def _unescape_bad_chars(path):
    """
    Decode illegal characters in mount point path.

    :param path: The path to unescape.
    :returns: The unescaped path.
    """

    def decode_byte(byte):
        return bytearray.fromhex(byte).decode("utf8")

    unescaped = ""
    i = 0
    while i < len(path):
        if path[i] == ".":
            if path[i + 1] == ".":
                unescaped = unescaped + "."
                i += 1
            else:
                unescaped = unescaped + decode_byte(path[i + 1 : i + 3])
                i += 2
        else:
            unescaped = unescaped + path[i]
        i += 1
    return unescaped


def encode_mount_point(mount_point):
    """
    Encode mount point paths.

    Encode a mount point for use in a snapshot name by replacing the path
    separator with '-'.
    """
    if mount_point == "/":
        return _MOUNT_SEPARATOR
    mount_point = _escape_bad_chars(mount_point)
    parts = [
        part.replace(_MOUNT_SEPARATOR, _ESCAPED_MOUNT_SEPARATOR)
        for part in mount_point.split(path_sep)
    ]
    return _MOUNT_SEPARATOR.join(parts)


def _split_mount_separators(mount_str):
    """
    Split string by mount separator handling escaped separators.

    Split the mount string ``mount_str`` at ``_MOUNT_SEPARATOR`` boundaries,
    into path components, ignoring ``_ESCAPED_MOUNT_SEPARATOR``.
    """
    result = []
    current = ""
    i = 0
    while i < len(mount_str):
        if mount_str[i] == _MOUNT_SEPARATOR:
            if i < len(mount_str) - 1 and mount_str[i + 1] == _MOUNT_SEPARATOR:
                current += _ESCAPED_MOUNT_SEPARATOR
                i += 1
            else:
                result.append(current)
                current = ""
        else:
            current += mount_str[i]
        i += 1

    result.append(current)  # Append the remaining segment
    return result


def decode_mount_point(mount_str):
    """
    Parse the mount point component of a snapshot name.
    """
    mount_str = _unescape_bad_chars(mount_str)
    mount_parts = _split_mount_separators(mount_str)
    unescaped_parts = [
        part.replace(_ESCAPED_MOUNT_SEPARATOR, _MOUNT_SEPARATOR) for part in mount_parts
    ]
    return path_sep.join(unescaped_parts)


def parse_snapshot_name(full_name, origin):
    """
    Attempt to parse a snapshot set snapshot name.

    Returns a tuple of (snapset_name, timestamp, mount_point) if ``full_name``
    is a valid snapset snapshot name, or ``None`` otherwise.
    """
    if not full_name.startswith(origin):
        return None
    base = full_name.removeprefix(origin + "-")
    if "_" not in base:
        return None
    fields = base.split("_", maxsplit=3)
    if len(fields) != 4:
        return None
    if fields[0] != "snapset":
        return None
    snapset_name = fields[1]
    timestamp = int(fields[2])
    mount_point = decode_mount_point(fields[3])
    # (snapset_name, timestamp, mount)
    return (snapset_name, timestamp, mount_point)


def _parse_proc_mounts_line(mount_line):
    """
    Parse ``/proc/mounts`` line.

    Parse a line of ``/proc/mounts`` data and return a tuple
    containing the device and mount point.
    """
    (device, mount_point, _) = mount_line.split(maxsplit=2)
    return (device, mount_point)


def device_from_mount_point(mount_point):
    """
    Convert a mount point path to a corresponding device path.

    Return the device corresponding to the file system mount point
    ``mount_point`` according to /proc/self/mounts.
    """
    # Ignore any trailing '/' in mount point path
    if mount_point != path_sep:
        mount_point = mount_point.rstrip(path_sep)
    with open(_PROC_MOUNTS, "r", encoding="utf8") as mounts:
        for line in mounts.readlines():
            (device, mount_path) = _parse_proc_mounts_line(line)
            if mount_path == mount_point:
                return device
    raise KeyError(f"Mount point {mount_point} not found")


def mount_point_space_used(mount_point):
    """
    Retrieve space usage for a mount point path.

    Return the space used on the file system mounted at ``mount_point``
    as a value in bytes.
    """
    if not mount_point or not ismount(mount_point):
        return -1
    stat = os.statvfs(mount_point)
    return (stat.f_blocks - stat.f_bfree) * stat.f_frsize


def format_snapshot_name(origin, snapset_name, timestamp, mount_point):
    """
    Format structured snapshot name.

    Format a snapshot name according to the snapshot manager naming format.
    """
    return SNAPSHOT_NAME_FORMAT % (origin, snapset_name, timestamp, mount_point)


__all__ = [
    "DEV_PREFIX",
    "DEV_MAPPER_PREFIX",
    "DMSETUP_CMD",
    "DMSETUP_SPLITNAME",
    "DMSETUP_INFO",
    "DMSETUP_LVM_SUBSYS",
    "DMSETUP_NO_HEADINGS",
    "DMSETUP_COLUMNS",
    "DMSETUP_FIELDS_VG_LV",
    "DMSETUP_FIELDS_UUID",
    "DMSETUP_REPORT_SEP",
    "encode_mount_point",
    "decode_mount_point",
    "parse_snapshot_name",
    "device_from_mount_point",
    "mount_point_space_used",
    "format_snapshot_name",
]



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
General constants.
"""
# isort: STDLIB
from enum import Enum

# isort: THIRDPARTY
from packaging.specifiers import Version

SERVICE = "org.storage.stratis3"
TOP_OBJECT = "/org/storage/stratis3"

MAXIMUM_STRATISD_VERSION = "4.0.0"
MINIMUM_STRATISD_VERSION = "3.7.0"
assert Version(MINIMUM_STRATISD_VERSION) < Version(MAXIMUM_STRATISD_VERSION)

REVISION = f"r{MINIMUM_STRATISD_VERSION.split('.')[1]}"

BLOCKDEV_INTERFACE = f"org.storage.stratis3.blockdev.{REVISION}"
FILESYSTEM_INTERFACE = f"org.storage.stratis3.filesystem.{REVISION}"
MANAGER_INTERFACE = f"org.storage.stratis3.Manager.{REVISION}"
POOL_INTERFACE = f"org.storage.stratis3.pool.{REVISION}"
REPORT_INTERFACE = f"org.storage.stratis3.Report.{REVISION}"


class IdType(Enum):
    """
    Whether the pool identifier is a UUID or a name.
    """

    UUID = 0
    NAME = 1


class Id:
    """
    Generic management of ids when either a UUID or name can be used.
    """

    def __init__(self, id_type, id_value):
        """
        Initialize the object.
        """
        self.id_type = id_type
        self.id_value = id_value

    def managed_objects_key(self):
        """
        Return key for managed objects.

        Precondition: the D-Bus property that identifies the Name or the
        Uuid will always be the same.
        """
        return {"Uuid" if self.id_type is IdType.UUID else "Name": self.dbus_value()}

    def is_uuid(self):
        """
        True if the id is a UUID, otherwise false.
        """
        return self.id_type == IdType.UUID

    def dbus_value(self):
        """
        Returns a string value for matching D-Bus things.
        """
        return self.id_value.hex if self.id_type is IdType.UUID else self.id_value

    def __str__(self):
        pool_id_type_str = "UUID" if self.id_type is IdType.UUID else "name"
        return f"{pool_id_type_str} {self.id_value}"



# Copyright Red Hat
#
# snapm/manager/_manager.py - Snapshot Manager
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
"""
Manager interface and plugin infrastructure.
"""
from subprocess import run, CalledProcessError
import logging
from time import time
from math import floor
from stat import S_ISBLK
from os import stat
from os.path import exists, ismount, normpath, samefile
import fnmatch
import inspect
import os

import snapm.manager.plugins
from snapm.manager.signals import suspend_signals
from snapm.manager.boot import (
    BootCache,
    create_snapset_boot_entry,
    delete_snapset_boot_entry,
    create_snapset_revert_entry,
    delete_snapset_revert_entry,
    check_boom_config,
)

from snapm import (
    SNAPM_DEBUG_MANAGER,
    SNAPM_VALID_NAME_CHARS,
    SnapmError,
    SnapmCalloutError,
    SnapmNoSpaceError,
    SnapmNoProviderError,
    SnapmExistsError,
    SnapmBusyError,
    SnapmPathError,
    SnapmNotFoundError,
    SnapmInvalidIdentifierError,
    SnapmPluginError,
    SnapmStateError,
    SnapmRecursionError,
    SnapmArgumentError,
    Selection,
    bool_to_yes_no,
    is_size_policy,
    SnapStatus,
    SnapshotSet,
    Snapshot,
)


_log = logging.getLogger(__name__)
_log.set_debug_mask(SNAPM_DEBUG_MANAGER)

_log_debug = _log.debug
_log_debug_manager = _log.debug_masked
_log_info = _log.info
_log_warn = _log.warning
_log_error = _log.error

JOURNALCTL_CMD = "journalctl"


class PluginRegistry(type):
    """
    Metaclass for Plugin classes.
    """

    plugins = []

    def __init__(cls, name, _bases, _attrs):
        super().__init__(name, _bases, _attrs)
        if name != "Plugin" and not name.startswith("_"):
            _log_debug("Loaded plugin %s version: %s", cls.name, cls.version)
            PluginRegistry.plugins.append(cls)


class Plugin(metaclass=PluginRegistry):
    """
    Abstract base class for snapshot manager plugins.
    """

    name = "plugin"
    version = "0.1.0"
    snapshot_class = Snapshot

    def __init__(self, logger):
        self.size_map = None
        self.logger = logger

    def _log_error(self, *args):
        """
        Log at error level.
        """
        self.logger.error(*args)

    def _log_warn(self, *args):
        """
        Log at warning level.
        """
        self.logger.warning(*args)

    def _log_info(self, *args):
        """
        Log at info level.
        """
        self.logger.info(*args)

    def _log_debug(self, *args):
        """
        Log at debug level.
        """
        self.logger.debug(*args)

    def info(self):
        """
        Return plugin name and version.
        """
        return {"name": self.name, "version": self.version}

    def start_transaction(self):
        """
        Begin a snapshot set creation transaction in this plugin.
        """
        self.size_map = {}

    def end_transaction(self):
        """
        End a snapshot set creation transaction in this plugin.
        """
        self.size_map = None

    def discover_snapshots(self):
        """
        Discover snapshots managed by this plugin class.

        Returns a list of objects that are a subclass of ``Snapshot``.
        :returns: A list of snapshots discovered by this plugin class.
        """
        raise NotImplementedError

    def can_snapshot(self, source):
        """
        Test whether this plugin can snapshot the specified mount point or
        block device.

        :param source: The block device or mount point path to test.
        :returns: ``True`` if this plugin can snapshot the file system mounted
                  at ``mount_point``, or ``False`` otherwise.
        """
        raise NotImplementedError

    def check_create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        """
        Perform pre-creation checks before creating a snapshot.

        :param origin: The origin volume for the snapshot.
        :param snapset_name: The name of the snapshot set to be created.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point path for this snapshot.
        :raises: ``SnapmNoSpaceError`` if there is insufficient free space to
                 create the snapshot.
        """
        raise NotImplementedError

    def create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        """
        Create a snapshot of ``origin`` in the snapset named ``snapset_name``.

        :param origin: The origin volume for the snapshot.
        :param snapset_name: The name of the snapshot set to be created.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point path for this snapshot.
        :raises: ``SnapmNoSpaceError`` if there is insufficient free space to
                 create the snapshot.
        """
        raise NotImplementedError

    def rename_snapshot(self, old_name, origin, snapset_name, timestamp, mount_point):
        """
        Rename the snapshot named ``old_name`` according to the provided
        snapshot field values.

        :param old_name: The original name of the snapshot to be renamed.
        :param origin: The origin volume for the snapshot.
        :param snapset_name: The new name of the snapshot set.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point of the snapshot.
        """
        raise NotImplementedError

    def check_resize_snapshot(self, name, origin, mount_point, size_policy):
        """
        Chcek whether this snapshot can be resized to the requested
        ``size_policy``. This method returns if the resize can be satisfied and
        raises an exception if not.

        :returns: None
        :raises: ``SnapmNoSpaceError`` if insufficient space is available to
        satisfy the requested size policy or ``SnapmPluginError`` if another
        reason prevents the snapshot from being resized.
        """
        raise NotImplementedError

    def resize_snapshot(self, name, origin, mount_point, size_policy):
        """
        Attempt to resize the snapshot ``name`` to the requested
        ``size_policy``. This method returns if the resize can be satisfied and
        raises an exception if not.

        :returns: None
        :raises: ``SnapmNoSpaceError`` if insufficient space is available to
        satisfy the requested size policy or ``SnapmPluginError`` if another
        reason prevents the snapshot from being resized.
        """
        raise NotImplementedError

    def check_revert_snapshot(self, name, origin):
        """
        Check whether this snapshot can be reverted or not. This method returns
        if the current snapshot can be reverted and raises an exception if not.

        :returns: None
        :raises: ``NotImplementedError`` if this plugin does not support the
        revert operation, ``SnapmBusyError`` if the snapshot is already in the
        process of being reverted to another snapshot state or
        ``SnapmPluginError`` if another reason prevents the snapshot from being
        merged.
        """
        raise NotImplementedError

    def revert_snapshot(self, name):
        """
        Request to revert a snapshot and revert the content of the origin
        volume to its state at the time of the snapshot.

        This may be deferred until the next device activation or mount
        operation for the respective volume.

        :param name: The name of the snapshot to revert.
        """
        raise NotImplementedError

    def delete_snapshot(self, name):
        """
        Delete the snapshot named ``name``

        :param name: The name of the snapshot to be removed.
        """
        raise NotImplementedError

    def activate_snapshot(self, name):
        """
        Activate the snapshot named ``name``

        :param name: The name of the snapshot to be activated.
        """
        raise NotImplementedError

    def deactivate_snapshot(self, name):
        """
        Deactivate the snapshot named ``name``

        :param name: The name of the snapshot to be deactivated.
        """
        raise NotImplementedError

    def set_autoactivate(self, name, auto=False):
        """
        Set the autoactivation state of the snapshot named ``name``.

        :param name: The name of the snapshot to be modified.
        :param auto: ``True`` to enable autoactivation or ``False`` otherwise.
        """
        raise NotImplementedError

    def origin_from_mount_point(self, mount_point):
        """
        Return a string representing the origin from a given mount point path.

        :param mount_point: The mount point path.
        """
        raise NotImplementedError


def find(file_pattern, top_dir, max_depth=None, path_pattern=None):
    """
    Generator function to find files recursively.
    Usage::

        for filename in find("*.properties", "/var/log/foobar"):
            print filename
    """
    if max_depth:
        base_depth = os.path.dirname(top_dir).count(os.path.sep)
        max_depth += base_depth

    for path, dirlist, filelist in os.walk(top_dir):
        if max_depth and path.count(os.path.sep) >= max_depth:
            del dirlist[:]

        if path_pattern and not fnmatch.fnmatch(path, path_pattern):
            continue

        for name in fnmatch.filter(filelist, file_pattern):
            yield os.path.join(path, name)


def _plugin_name(path):
    """
    Returns the plugin module name given the path.
    """
    base = os.path.basename(path)
    name, _ = os.path.splitext(base)
    return name


def _get_plugins_from_list(pluglist):
    """
    Get list of plugin names from file list.

    :param pluglist: list of candidate plugin files.
    """
    plugins = [
        _plugin_name(plugin)
        for plugin in pluglist
        if "__init__" not in plugin and plugin.endswith(".py")
    ]
    plugins.sort()
    return plugins


def _find_plugins_in_dir(path):
    """
    Find possible plugin files in ``path``.

    :param path: The search directory for plugin files.
    """
    _log_debug_manager("Finding plugins in %s", path)
    if os.path.exists(path):
        py_files = list(find("[a-zA-Z]*.py", path))
        pnames = _get_plugins_from_list(py_files)
        _log_debug_manager("Found plugin modules: %s", ", ".join(pnames))
        if pnames:
            return pnames
    return []


class ImporterHelper:
    """
    Provides a list of modules that can be imported in a package.

    Importable modules are located along the module __path__ list and modules
    are files that end in .py.
    """

    def __init__(self, package):
        """
        package is a package module

        import my.package.module
        helper = ImporterHelper(my.package.module)
        """
        self.package = package

    def get_modules(self):
        """
        Get list of importable modules.

        Returns the list of importable modules in the configured python
        package.
        """
        plugins = []
        for path in self.package.__path__:
            if os.path.isdir(path):
                plugins.extend(_find_plugins_in_dir(path))

        return plugins


def import_module(module_fqname, superclasses=None):
    """
    Import a single module.

    Imports the module module_fqname and returns a list of defined classes
    from that module. If superclasses is defined then the classes returned will
    be subclasses of the specified superclass or superclasses. If superclasses
    is plural it must be a tuple of classes.
    """
    module_name = module_fqname.rpartition(".")[-1]
    try:
        module = __import__(module_fqname, globals(), locals(), [module_name])
    # pylint: disable=broad-except
    except Exception as err:
        _log_error("Error importing %s plugin module: %s", module_fqname, err)
        return []
    modules = [
        class_
        for cname, class_ in inspect.getmembers(module, inspect.isclass)
        if class_.__module__ == module_fqname
    ]
    if superclasses:
        modules = [m for m in modules if issubclass(m, superclasses)]

    return modules


def import_plugin(name, superclasses=None):
    """
    Import name as a module and return a list of all classes defined in that
    module. superclasses should be a tuple of valid superclasses to import,
    this defaults to (Plugin,).
    """
    plugin_fqname = f"snapm.manager.plugins.{name}"
    if not superclasses:
        superclasses = (Plugin,)
    _log_debug("Importing plugin module %s", plugin_fqname)
    return import_module(plugin_fqname, superclasses)


def load_plugins():
    """
    Attempt to load plugin modules.
    """
    helper = ImporterHelper(snapm.manager.plugins)
    plugins = helper.get_modules()
    _log_debug(
        "Importing %d modules from %s", len(plugins), snapm.manager.plugins.__name__
    )
    for plug in plugins:
        plugbase, _ = os.path.splitext(plug)
        if not plugbase.startswith("_"):
            import_plugin(plugbase)


def select_snapshot_set(select, snapshot_set):
    """
    Test SnapshotSet against Selection criteria.

    Test the supplied ``SnapshotSet`` against the selection criteria
    in ``s`` and return ``True`` if it passes, or ``False``
    otherwise.

    :param s: The selection criteria
    :param be: The SnapshotSet to test
    :rtype: bool
    :returns: True if SnapshotSet passes selection or ``False``
              otherwise.
    """
    if select.name and select.name != snapshot_set.name:
        return False
    if select.uuid and select.uuid != snapshot_set.uuid:
        return False
    if select.timestamp and select.timestamp != snapshot_set.timestamp:
        return False
    if select.nr_snapshots and select.nr_snapshots != snapshot_set.nr_snapshots:
        return False
    if select.mount_points:
        s_mount_points = sorted(select.mount_points)
        mount_points = sorted(snapshot_set.mount_points)
        if s_mount_points != mount_points:
            return False

    return True


def select_snapshot(select, snapshot):
    """
    Test SnapshotSet against Selection criteria.

    Test the supplied ``SnapshotSet`` against the selection criteria
    in ``s`` and return ``True`` if it passes, or ``False``
    otherwise.

    :param s: The selection criteria
    :param be: The SnapshotSet to test
    :rtype: bool
    :returns: True if SnapshotSet passes selection or ``False``
              otherwise.
    """
    if not select_snapshot_set(select, snapshot.snapshot_set):
        return False
    if select.snapshot_uuid and select.snapshot_uuid != snapshot.uuid:
        return False
    if select.snapshot_name and select.snapshot_name != snapshot.name:
        return False

    return True


def _suspend_journal():
    """
    Suspend journal writes to /var before creating snapshots.
    """
    try:
        run([JOURNALCTL_CMD, "--flush"], check=True)
        run([JOURNALCTL_CMD, "--relinquish-var"], check=True)
    except CalledProcessError as err:  # pragma: no cover
        raise SnapmCalloutError(
            f"Error calling journalctl to flush journal: {err}"
        ) from err


def _resume_journal():
    """
    Resume journal writes to /var after creating snapshots.
    """
    try:
        run([JOURNALCTL_CMD, "--flush"], check=True)
    except CalledProcessError as err:  # pragma: no cover
        raise SnapmCalloutError(
            f"Error calling journalctl to flush journal: {err}"
        ) from err


def _parse_source_specs(source_specs, default_size_policy):
    """
    Parse and normalize source paths and size policies.

    :param source_specs: A list of mount point or block device paths and
                         optional size policy strings.
    :returns: A tuple (sources, size_policies) containing a list of
              normalized mount point and block device paths and a dictionary
              mapping source paths to size policies.
    """
    sources = []
    size_policies = {}

    # Parse size policies and normalise mount paths
    for spec in source_specs:
        if ":" in spec:
            (source, policy) = spec.rsplit(":", maxsplit=1)
            if not is_size_policy(policy):
                source = f"{source}:{policy}"
                policy = default_size_policy
        else:
            (source, policy) = (spec, default_size_policy)
        source = normpath(source)
        sources.append(source)
        size_policies[source] = policy
    return (sources, size_policies)


def _check_revert_snapshot_set(snapset):
    """
    Check that all snapshots in a snapshot set can be reverted.
    Returns if all checks pass or raises an exception otherwise.

    :returns: None
    :raises: ``NotImplementedError``, ``SnapmBusyError`` or
             ``SnapmPluginError``
    """
    if snapset.status == SnapStatus.INVALID:
        raise SnapmStateError(
            f"Cannot revert snapset {snapset.name} with invalid snapshots"
        )

    # Check for ability to revert all snapshots in set
    for snapshot in snapset.snapshots:
        try:
            snapshot.check_revert()
        except NotImplementedError as err:
            _log_error(
                "Snapshot provider %s does not support revert",
                snapshot.provider.name,
            )
            raise SnapmPluginError from err
        except SnapmBusyError as err:
            _log_error(
                "Revert already in progress for snapshot origin %s",
                snapshot.origin,
            )
            raise SnapmBusyError(
                f"A revert has already started for snapshot set {snapset.name}"
            ) from err
        except SnapmPluginError as err:
            _log_error(
                "Revert prechecks failed for snapshot %s (%s)",
                snapshot.name,
                snapshot.provider.name,
            )
            raise err


def _check_snapset_status(snapset, operation):
    """
    Check that a snapshot set status is not ``SnapStatus.INVALID`` or
    ``SnapStatus.REVERTING`` before carrying out ``operation``.
    """
    if snapset.status in (SnapStatus.INVALID, SnapStatus.REVERTING):
        if snapset.status == SnapStatus.INVALID:
            _log_error("Cannot operate on invalid snapshot set '%s'", snapset.name)
        if snapset.status == SnapStatus.REVERTING:
            _log_error("Cannot operate on reverting snapshot set '%s'", snapset.name)
        raise SnapmStateError(
            f"Failed to {operation} snapset '{snapset.name}': status is '{snapset.status}'"
        )


def _find_mount_point_for_devpath(devpath):
    """
    Return the first mount point found in /proc/mounts that corresponds to
    ``device``, or the empty string if no mount point can be found.
    """
    with open("/proc/mounts", "r", encoding="utf8") as mounts:
        for line in mounts:
            fields = line.split(" ")
            if exists(fields[0]):
                if samefile(devpath, fields[0]):
                    return fields[1]
    return ""


class Manager:
    """
    Snapshot Manager high level interface.
    """

    plugins = []
    snapshot_sets = []
    by_name = {}
    by_uuid = {}

    @suspend_signals
    def __init__(self):
        self.plugins = []
        self.snapshot_sets = []
        self.by_name = {}
        self.by_uuid = {}
        check_boom_config()
        self._boot_cache = BootCache()
        load_plugins()
        for plugin_class in PluginRegistry.plugins:
            try:
                plugin = plugin_class(_log)
                self.plugins.append(plugin)
            except SnapmNotFoundError as err:
                _log_debug(
                    "Plugin dependencies missing: %s (%s), skipping.",
                    plugin_class.__name__,
                    err,
                )
            except SnapmPluginError as err:
                _log_error("Disabling plugin %s: %s", plugin_class.__name__, err)
        self.discover_snapshot_sets()

    def _find_and_verify_plugins(
        self, sources, size_policies, _requested_provider=None
    ):
        """
        Find snapshot provider plugins for each source in ``sources``
        and verify that a provider exists for each source present.

        :param sources: A list of source mount point or block device paths.
        :param size_policies: A dictionary mapping sources to size policies.
        :returns: A dictionary mapping sources to plugins.
        """
        # Initialise provider mapping.
        provider_map = {k: None for k in sources}

        # Find provider plugins for sources
        for source in sources:
            if not exists(source):
                _log_error("No such file or directory: %s", source)
                raise SnapmNotFoundError(f"Source path '{source}' does not exist")

            _log_debug(
                "Probing plugins for %s with size policy %s",
                source,
                size_policies[source],
            )

            if not ismount(source) and not S_ISBLK(stat(source).st_mode):
                raise SnapmPathError(
                    f"Path '{source}' is not a block device or mount point"
                )

            for plugin in self.plugins:
                if plugin.can_snapshot(source):
                    provider_map[source] = plugin

        # Verify each mount point has a provider plugin
        for source in provider_map:
            if provider_map[source] is None:
                raise SnapmNoProviderError(
                    f"Could not find snapshot provider for {source}"
                )
        return provider_map

    def _snapset_from_name_or_uuid(self, name=None, uuid=None):
        """
        Look a snapshot set up by ``name`` or ``uuid``. Returns a
        ``SnapshotSet`` correponding to either ``name`` or ``uuid`,
        or raises ``SnapmError`` on error.

        :param name: The name of the snapshot set to look up.
        :param uuid: The UUID of the snapshot set to look up.
        :returns: A ``SnapshotSet`` corresponding to the given name or UUID.
        :raises: ``SnapmNotFoundError`` is the name or UUID cannot be found.
                 ``SnapmInvalidIdentifierError`` if the name and UUID do not
                 match.
        """
        if name is not None:
            if name not in self.by_name:
                raise SnapmNotFoundError(f"Could not find snapshot set named {name}")
            snapset = self.by_name[name]
        elif uuid is not None:
            if uuid not in self.by_uuid:
                raise SnapmNotFoundError(
                    f"Could not find snapshot set with uuid {uuid}"
                )
            snapset = self.by_uuid[uuid]
        else:
            raise SnapmNotFoundError("A snapshot set name or UUID is required")

        if name and uuid:
            if self.by_name[name] != self.by_uuid[uuid]:
                raise SnapmInvalidIdentifierError(
                    f"Conflicting name and UUID: {str(uuid)} does not match '{name}'"
                )

        return snapset

    def _check_recursion(self, origins):
        """
        Verify that each entry in  ``origins`` corresponds to a device that is
        not a snapshot belonging to another snapshot set.

        :param origins: A list of origin devices to check.
        :raises: ``SnapmRecursionError`` if an origin device is a snapshot.
        """
        snapshot_devices = [
            snapshot.devpath
            for snapset in self.snapshot_sets
            for snapshot in snapset.snapshots
        ]
        for source, device in origins.items():
            if device in snapshot_devices:
                raise SnapmRecursionError(
                    "Snapshots of snapshots are not supported: "
                    f"{source} corresponds to snapshot device {device}"
                )

    def discover_snapshot_sets(self):
        """
        Discover snapshot sets by calling into each plugin to find
        individual snapshots and then aggregating them together into
        snapshot sets.

        Initialises the ``snapshot_sets``, ``by_name`` and ``by_uuid`` members
        with the discovered snapshot sets.
        """
        self.snapshot_sets = []
        self.by_name = {}
        self.by_uuid = {}
        snapshots = []
        snapset_names = set()
        self._boot_cache.refresh_cache()
        _log_debug("Discovering snapshot sets for %s plugins", len(self.plugins))
        for plugin in self.plugins:
            snapshots.extend(plugin.discover_snapshots())
        _log_debug("Discovered %s managed snapshots", len(snapshots))
        for snapshot in snapshots:
            snapset_names.add(snapshot.snapset_name)
        for snapset_name in snapset_names:
            set_snapshots = [
                snap for snap in snapshots if snap.snapset_name == snapset_name
            ]
            set_timestamp = set_snapshots[0].timestamp
            for snap in set_snapshots:
                if snap.timestamp != set_timestamp:
                    _log_warn(
                        "Snapshot set '%s' has inconsistent timestamps", snapset_name
                    )
                    continue

            snapset = SnapshotSet(snapset_name, set_timestamp, set_snapshots)

            # Associate snapset with boot entry if present
            if snapset.name in self._boot_cache.entry_cache:
                snapset.boot_entry = self._boot_cache.entry_cache[snapset.name]
            elif str(snapset.uuid) in self._boot_cache.entry_cache:
                snapset.boot_entry = self._boot_cache.entry_cache[str(snapset.uuid)]

            # Associate snapset with revert entry if present
            if snapset.name in self._boot_cache.revert_cache:
                snapset.revert_entry = self._boot_cache.revert_cache[snapset.name]
            elif str(snapset.uuid) in self._boot_cache.revert_cache:
                snapset.revert_entry = self._boot_cache.revert_cache[str(snapset.uuid)]

            self.snapshot_sets.append(snapset)
            self.by_name[snapset.name] = snapset
            self.by_uuid[snapset.uuid] = snapset
            for snapshot in snapset.snapshots:
                snapshot.snapshot_set = snapset

        self.snapshot_sets.sort(key=lambda ss: ss.name)
        _log_debug("Discovered %d snapshot sets", len(self.snapshot_sets))

    def find_snapshot_sets(self, selection=None):
        """
        Find snapshot sets matching selection criteria.

        :param selection: Selection criteria to apply.
        :returns: A list of matching ``SnapshotSet`` objects.
        """
        matches = []

        # Use null selection criteria if unspecified
        selection = selection if selection else Selection()

        selection.check_valid_selection(snapshot_set=True)

        _log_debug("Finding snapshot sets for %s", repr(selection))

        for snapset in self.snapshot_sets:
            if select_snapshot_set(selection, snapset):
                matches.append(snapset)

        _log_debug("Found %d snapshot sets", len(matches))
        return matches

    def find_snapshots(self, selection=None):
        """
        Find snapshots matching selection criteria.

        :param selection: Selection criteria to apply.
        :returns: A list of matching ``Snapshot`` objects.
        """
        matches = []

        # Use null selection criteria if unspecified
        selection = selection if selection else Selection()

        selection.check_valid_selection(snapshot=True, snapshot_set=True)

        _log_debug("Finding snapshots for %s", repr(selection))

        for snapset in self.snapshot_sets:
            for snapshot in snapset.snapshots:
                if select_snapshot(selection, snapshot):
                    matches.append(snapshot)

        _log_debug("Found %d snapshots", len(matches))
        return matches

    def _validate_snapset_name(self, name):
        """
        Validate a snapshot set name.

        Returns if ``name`` is a valid snapshot set name or raises an
        appropriate exception otherwise.

        :param name: The snapshot set name to validate.
        :raises: ``SnapmExistsError`` if the name is already in use, or
                 ``SnapmInvalidIdentifierError`` if the name fails validation.
        """
        if name in self.by_name:
            raise SnapmExistsError(f"Snapshot set named '{name}' already exists")
        for char in name:
            # Underscore is specifically disallowed in snapset names.
            if char == "_" or char not in SNAPM_VALID_NAME_CHARS:
                raise SnapmInvalidIdentifierError(
                    f"Snapshot set name cannot include '{char}'"
                )

    # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    @suspend_signals
    def create_snapshot_set(
        self, name, source_specs, default_size_policy=None, boot=False, revert=False
    ):
        """
        Create a snapshot set of the supplied mount points with the name
        ``name``.

        :param name: The name of the snapshot set.
        :param source_specs: A list of mount point and block device paths to
                             include in the set.
        :param default_size_policy: A default size policy to use for the set.
        :raises: ``SnapmExistsError`` if the name is already in use, or
                 ``SnapmInvalidIdentifierError`` if the name fails validation.
        """
        self._validate_snapset_name(name)

        # Parse size policies and normalise mount paths
        (sources, size_policies) = _parse_source_specs(
            source_specs, default_size_policy
        )

        # Initialise provider mapping.
        provider_map = self._find_and_verify_plugins(sources, size_policies, None)

        for provider in set(provider_map.values()):
            provider.start_transaction()

        timestamp = floor(time())
        origins = {}
        mounts = {}

        for source in provider_map:
            if S_ISBLK(stat(source).st_mode):
                mounts[source] = _find_mount_point_for_devpath(source)
                origins[source] = source
                mount = mounts[source]
                if mount in provider_map:
                    raise SnapmInvalidIdentifierError(
                        f"Duplicate snapshot source {source} already added to {name} as {mount}"
                    )
            else:
                mount = source
                origins[source] = provider_map[source].origin_from_mount_point(mount)

            try:
                provider_map[source].check_create_snapshot(
                    origins[source], name, timestamp, mount, size_policies[source]
                )
            except SnapmInvalidIdentifierError as err:
                _log_error(
                    "Error creating %s snapshot: %s", provider_map[source].name, err
                )
                raise SnapmInvalidIdentifierError(
                    f"Snapset name {name} too long"
                ) from err
            except SnapmNoSpaceError as err:
                _log_error(
                    "Error creating %s snapshot: %s", provider_map[source].name, err
                )
                raise SnapmNoSpaceError(
                    f"Insufficient free space for snapshot set {name}"
                ) from err

        self._check_recursion(origins)

        _suspend_journal()

        snapshots = []
        for source in provider_map:
            if S_ISBLK(stat(source).st_mode):
                mount = mounts[source]
            else:
                mount = source
            try:
                snapshots.append(
                    provider_map[source].create_snapshot(
                        origins[source], name, timestamp, mount, size_policies[source]
                    )
                )
            except SnapmError as err:
                _log_error("Error creating snapshot set member %s: %s", name, err)
                _resume_journal()
                for snapshot in snapshots:
                    snapshot.delete()
                raise SnapmPluginError(
                    f"Could not create all snapshots for set {name}"
                ) from err

        _resume_journal()

        for provider in set(provider_map.values()):
            provider.end_transaction()

        snapset = SnapshotSet(name, timestamp, snapshots)

        if boot or revert:
            _log_info(
                "Autoactivation required for bootable snapshot set '%s'", snapset.name
            )
            self._set_autoactivate(snapset, auto=True)
            snapset.activate()

        if boot:
            try:
                create_snapset_boot_entry(snapset)
                self._boot_cache.refresh_cache()
            except (OSError, ValueError) as err:
                _log_error("Failed to create snapshot set boot entry: %s", err)
                snapset.delete()
                raise SnapmCalloutError from err

        if revert:
            try:
                create_snapset_revert_entry(snapset)
                self._boot_cache.refresh_cache()
            except (OSError, ValueError) as err:
                _log_error("Failed to create snapshot set revert boot entry: %s", err)
                snapset.delete()
                raise SnapmCalloutError from err

        self.by_name[snapset.name] = snapset
        self.by_uuid[snapset.uuid] = snapset
        self.snapshot_sets.append(snapset)
        return snapset

    @suspend_signals
    def rename_snapshot_set(self, old_name, new_name):
        """
        Rename snapshot set ``old_name`` as ``new_name``.

        :param old_name: The name of the snapshot set to be renamed.
        :param new_name: The new name of the snapshot set.
        :raises: ``SnapmExistsError`` if the name is already in use, or
                 ``SnapmInvalidIdentifierError`` if the name fails validation.
        """
        if old_name not in self.by_name:
            raise SnapmNotFoundError(f"Cannot find snapshot set named {old_name}")

        self._validate_snapset_name(new_name)

        snapset = self.by_name[old_name]
        _check_snapset_status(snapset, "rename")

        # Remove references to old set
        self.by_name.pop(snapset.name)
        self.by_uuid.pop(snapset.uuid)
        self.snapshot_sets.remove(snapset)

        try:
            snapset.rename(new_name)
        except SnapmError as err:
            self.by_name[snapset.name] = snapset
            self.by_uuid[snapset.uuid] = snapset
            self.snapshot_sets.append(snapset)
            raise err

        self.by_name[snapset.name] = snapset
        self.by_uuid[snapset.uuid] = snapset
        self.snapshot_sets.append(snapset)
        return snapset

    @suspend_signals
    def delete_snapshot_sets(self, selection):
        """
        Remove snapshot sets matching selection criteria ``selection``.

        :param selection: Selection criteria for snapshot sets to remove.
        """
        sets = self.find_snapshot_sets(selection=selection)
        deleted = 0
        if not sets:
            raise SnapmNotFoundError(
                f"Could not find snapshot sets matching {selection}"
            )
        for snapset in sets:
            delete_snapset_boot_entry(snapset)
            delete_snapset_revert_entry(snapset)
            snapset.delete()
            self.snapshot_sets.remove(snapset)
            self.by_name.pop(snapset.name)
            self.by_uuid.pop(snapset.uuid)
            deleted += 1
        self._boot_cache.refresh_cache()
        return deleted

    # pylint: disable=too-many-branches
    @suspend_signals
    def resize_snapshot_set(
        self, source_specs, name=None, uuid=None, default_size_policy=None
    ):
        """
        Resize snapshot set named ``name`` or having UUID ``uuid``.

        Request to resize each snapshot included in ``source_specs``
        according to the given size policy, or apply ``default_size_policy``
        if set.

        :param name: The name of the snapshot set to resize.
        :param uuid: The UUID of the snapshot set to resize.
        :param source_specs: A list of mount points and optional size
                                  policies.
        :param default_size_policy: A default size policy to apply to the
                                    resize.
        """
        snapset = self._snapset_from_name_or_uuid(name=name, uuid=uuid)

        if source_specs:
            # Parse size policies and normalise mount paths
            (sources, size_policies) = _parse_source_specs(
                source_specs, default_size_policy
            )
        else:
            sources = [snapshot.source for snapshot in snapset.snapshots]
            size_policies = {source: default_size_policy for source in sources}

        snapset.resize(sources, size_policies)

    @suspend_signals
    def revert_snapshot_set(self, name=None, uuid=None):
        """
        Revert snapshot set named ``name`` or having UUID ``uuid``.

        Request to revert each snapshot origin within each snapshot set
        to the state at the time the snapshot was taken.

        :param name: The name of the snapshot set to revert.
        :param uuid: The UUID of the snapshot set to revert.
        """
        snapset = self._snapset_from_name_or_uuid(name=name, uuid=uuid)

        _check_revert_snapshot_set(snapset)

        # Snapshot boot entry becomes invalid as soon as revert is initiated.
        delete_snapset_boot_entry(snapset)

        snapset.revert()

        self._boot_cache.refresh_cache()
        return snapset

    def revert_snapshot_sets(self, selection):
        """
        Revert snapshot sets matching selection criteria ``selection``.

        Request to revert each snapshot origin within each snapshot set
        to the state at the time the snapshot was taken.

        :param selection: Selection criteria for snapshot sets to revert.
        """
        sets = self.find_snapshot_sets(selection=selection)
        reverted = 0
        if not sets:
            raise SnapmNotFoundError(
                f"Could not find snapshot sets matching {selection}"
            )
        for snapset in sets:
            self.revert_snapshot_set(name=snapset.name)
            reverted += 1
        return reverted

    @suspend_signals
    def activate_snapshot_sets(self, selection):
        """
        Activate snapshot sets matching selection criteria ``selection``.

        :param selection: Selection criteria for snapshot sets to activate.
        """
        sets = self.find_snapshot_sets(selection=selection)
        activated = 0
        if not sets:
            raise SnapmNotFoundError(
                f"Could not find snapshot sets matching {selection}"
            )
        for snapset in sets:
            _check_snapset_status(snapset, "activate")
            snapset.activate()
            activated += 1
        return activated

    @suspend_signals
    def deactivate_snapshot_sets(self, selection):
        """
        Deactivate snapshot sets matching selection criteria ``selection``.

        :param selection: Selection criteria for snapshot sets to deactivate.
        """
        sets = self.find_snapshot_sets(selection=selection)
        deactivated = 0
        if not sets:
            raise SnapmNotFoundError(
                f"Could not find snapshot sets matching {selection}"
            )
        for snapset in sets:
            _check_snapset_status(snapset, "deactivate")
            snapset.deactivate()
            deactivated += 1
        return deactivated

    def _set_autoactivate(self, snapset, auto=False):
        """
        Set autoactivation for ``snapset``.

        :param snapset: The ``SnapshotSet`` object to operate on.
        :param auto:  ``True`` to enable autoactivation or ``False`` otherwise.
        """
        _check_snapset_status(snapset, "set autoactivate status for")
        state = bool_to_yes_no(auto)
        _log_info("Setting autoactivation=%s for snapshot set %s", state, snapset.name)
        snapset.autoactivate = auto

    @suspend_signals
    def set_autoactivate(self, selection, auto=False):
        """
        Set autoactivation state for snapshot sets matching selection criteria
        ``selection``.

        :param selection: Selection criteria for snapshot sets to set
                          autoactivation.
        :param auto: ``True`` to enable autoactivation or ``False`` otherwise.
        """
        sets = self.find_snapshot_sets(selection=selection)
        changed = 0
        if not sets:
            raise SnapmNotFoundError(
                f"Could not find snapshot sets matching {selection}"
            )
        for snapset in sets:
            self._set_autoactivate(snapset, auto=auto)
            changed += 1
        return changed

    @suspend_signals
    def split_snapshot_set(self, name, new_name, source_specs):
        """
        Split the snapshot set named `name` into two snapshot sets, with
        `new_name` containing the listed `sources` and `name` containing all
        remaining snapshots.

        If `new_name` is `None` the listed snapshot are removed from `name`
        and permanently deleted. This operation cannot be undone.

        :param name: The name of the snapshot set to split
        :param new_name: The name of the newly split off snapshot set
        :param sources: The list of sources to include in the new snapshot set
        """
        if name not in self.by_name:
            raise SnapmNotFoundError(f"Cannot find snapshot set named {name}")

        if new_name:
            self._validate_snapset_name(new_name)

        op = f"{'split' if new_name else 'prune'}"

        _log_info("Attempting to %s snapshot set '%s'", op, name)

        snapset = self.by_name[name]
        _check_snapset_status(snapset, op)

        if not source_specs:
            raise SnapmArgumentError(
                f"SnapshotSet {op} requires at least one source argument"
            )

        # Parse size policies and normalise mount paths
        (sources, size_policies) = _parse_source_specs(source_specs, None)

        if any(size_policies[source] is not None for source in sources):
            err_sources = ", ".join(
                [source for source in source_specs if ":" in source]
            )
            raise SnapmArgumentError(
                f"SnapshotSet {op} does not support size policies: {err_sources}"
            )

        missing_sources = [
            source for source in sources if source not in snapset.sources
        ]

        if missing_sources:
            raise SnapmNotFoundError(
                f"SnapshotSet '{snapset.name}' does not contain {', '.join(missing_sources)}"
            )

        timestamp = snapset.timestamp

        split_snapshots = [
            snapshot for snapshot in snapset.snapshots if snapshot.source in sources
        ]

        keep_snapshots = [
            snapshot for snapshot in snapset.snapshots if snapshot.source not in sources
        ]

        if not keep_snapshots:
            op_str = "Splitting" if new_name else "Pruning"
            raise SnapmArgumentError(
                f"{op_str} {', '.join((snap.source for snap in split_snapshots))} "
                f"from {snapset.name} would leave {snapset.name} empty"
            )

        _log_debug(
            "Splitting snapshot%s %s from snapshot set '%s'",
            "s" if len(split_snapshots) > 1 else "",
            ", ".join((snap.source for snap in split_snapshots)),
            snapset.name,
        )

        self.by_name.pop(snapset.name)
        self.by_uuid.pop(snapset.uuid)
        self.snapshot_sets.remove(snapset)

        # Build new SnapshotSet with old name to start with, then use
        # SnapshotSet.rename() to propagate new_name to members.
        new_set = SnapshotSet(snapset.name, timestamp, split_snapshots)

        # Update original snapshot set and re-add to lookup tables
        orig_set = SnapshotSet(snapset.name, timestamp, keep_snapshots)
        self.by_name[orig_set.name] = orig_set
        self.by_uuid[orig_set.uuid] = orig_set
        self.snapshot_sets.append(orig_set)

        if new_name:
            # Attempt to rename members to new_name
            try:
                new_set.rename(new_name)
            except SnapmError as err:
                self.by_name[snapset.name] = snapset
                self.by_uuid[snapset.uuid] = snapset
                self.snapshot_sets.append(snapset)
                raise err
        else:
            try:
                # Delete the pruned members
                new_set.delete()
            except SnapmError as err:
                _log_error(
                    "Failed to clean up pruned snapshot set members: %s (%s)",
                    err,
                    ", ".join(snapshot.name for snapshot in new_set.snapshots),
                )
            return orig_set

        # Add new snapshot set to lookup tables
        self.by_name[new_set.name] = new_set
        self.by_uuid[new_set.uuid] = new_set
        self.snapshot_sets.append(new_set)

        return new_set

    @suspend_signals
    def create_snapshot_set_boot_entry(self, name=None, uuid=None):
        """
        Create a snapshot boot entry for the specified snapshot set.

        :param name: The name of the snapshot set.
        :param uuid: The UUID of the snapshot set.
        """
        snapset = self._snapset_from_name_or_uuid(name=name, uuid=uuid)

        snapset.autoactivate = True
        if not snapset.autoactivate:
            raise SnapmPluginError(
                "Could not enable autoactivation for all snapshots in snapshot "
                f"set {snapset.name}"
            )
        if snapset.boot_entry is not None:
            raise SnapmExistsError(
                f"Boot entry already associated with snapshot set {snapset.name}"
            )
        create_snapset_boot_entry(snapset)
        self._boot_cache.refresh_cache()

    @suspend_signals
    def create_snapshot_set_revert_entry(self, name=None, uuid=None):
        """
        Create a revert boot entry for the specified snapshot set.

        :param name: The name of the snapshot set.
        :param uuid: The UUID of the snapshot set.
        """
        snapset = self._snapset_from_name_or_uuid(name=name, uuid=uuid)

        if snapset.revert_entry is not None:
            raise SnapmExistsError(
                f"Revert entry already associated with snapshot set {snapset.name}"
            )
        create_snapset_revert_entry(snapset)
        self._boot_cache.refresh_cache()


__all__ = [
    "Plugin",
    "Manager",
]



# Copyright Red Hat
#
# snapm/__init__.py - Snapshot Manager package initialisation
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
"""
Snapm top-level package.
"""
from ._snapm import *  # noqa: F401, F403
from ._snapm import __all__  # noqa: F401

__version__ = "0.4.1"



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Environment functions.
"""

from ._errors import StratisCliEnvironmentError


def get_timeout(value):
    """
    Turn an input str or int into a float timeout value.

    :param value: the input str or int
    :type value: str or int
    :raises StratisCliEnvironmentError:
    :returns: float
    """

    maximum_dbus_timeout_ms = 1073741823

    # Ensure the input str is not a float
    if isinstance(value, float):
        raise StratisCliEnvironmentError(
            "The timeout value provided is a float; it should be an integer."
        )

    try:
        timeout_int = int(value)

    except ValueError as err:
        raise StratisCliEnvironmentError(
            "The timeout value provided is not an integer."
        ) from err

    # Ensure the integer is not too small
    if timeout_int < -1:
        raise StratisCliEnvironmentError(
            "The timeout value provided is smaller than the smallest acceptable value, -1."
        )

    # Ensure the integer is not too large
    if timeout_int > maximum_dbus_timeout_ms:
        raise StratisCliEnvironmentError(
            f"The timeout value provided exceeds the largest acceptable value, "
            f"{maximum_dbus_timeout_ms}."
        )

    # Convert from milliseconds to seconds
    return timeout_int / 1000



# Copyright Red Hat
#
# tests/test_lvm2.py - LVM2 plugin unit tests
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
import unittest
import logging
import os.path
import os

log = logging.getLogger()

import snapm.manager.plugins.lvm2 as lvm2
from snapm import SnapmCalloutError


class Lvm2Tests(unittest.TestCase):
    """Test lvm2 plugin functions"""

    _old_path = None

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)
        def cleanup():
            log.debug("Cleaning up (%s)", self._testMethodName)
            if hasattr(self, "_old_path"):
                os.environ["PATH"] = self._old_path

        self.addCleanup(cleanup)

        bin_path = os.path.abspath("tests/bin")
        cur_path = os.environ["PATH"]
        self._old_path = cur_path
        os.environ["PATH"] = bin_path + os.pathsep + cur_path

    def test_is_lvm_device(self):
        devs = {
            "/dev/mapper/fedora-home": True,
            "/dev/mapper/fedora-root": True,
            "/dev/mapper/fedora-var": True,
            "/dev/mapper/stratis-1-1c7c941a2dba4eb78d57d3fb01aacc61-thin-fs-202ea1667fa54123bd24f0c353b9914c": False,
            "/dev/mapper/mpatha": False,
        }
        for dev in devs.keys():
            self.assertEqual(lvm2.is_lvm_device(dev), devs[dev])

    def test_vg_lv_from_device_path(self):
        lvm2cow = lvm2.Lvm2Cow(log)
        devs = {
            "/dev/mapper/fedora-home": ("fedora", "home"),
            "/dev/mapper/fedora-root": ("fedora", "root"),
            "/dev/mapper/fedora-var": ("fedora", "var"),
            "/dev/mapper/appdata-datavol1--test": ("appdata", "datavol1-test"),
            "/dev/mapper/vg_test0-lv_test0": ("vg_test0", "lv_test0"),
            "/dev/mapper/notadev": None,
            "/dev/fedora/home": ("fedora", "home"),
            "/dev/fedora/root": ("fedora", "root"),
            "/dev/fedora/var": ("fedora", "var"),
            "/dev/appdata/datavol1-test": ("appdata", "datavol1-test"),
            "/dev/vg_test0/lv_test0": ("vg_test0", "lv_test0"),
            "/dev/not/a/dev": None,
        }
        for dev in devs.keys():
            if devs[dev] is not None:
                self.assertEqual(lvm2cow.vg_lv_from_device_path(dev), devs[dev])
            else:
                with self.assertRaises(SnapmCalloutError) as cm:
                    lvm2cow.vg_lv_from_device_path(dev)

    def test_vg_lv_from_origin(self):
        devs = {
            "/dev/fedora/root": ("fedora", "root"),
            "/dev/fedora/home": ("fedora", "home"),
            "/dev/rhel/var": ("rhel", "var"),
            "/dev/vg00/lvol00": ("vg00", "lvol00"),
        }
        for dev in devs.keys():
            self.assertEqual(lvm2.vg_lv_from_origin(dev), devs[dev])

    def test_pool_name_from_vg_lv(self):
        lvm2thin = lvm2.Lvm2Thin(log)
        devs = {
            "fedora/srv": "pool0",
            "fedora/home": "",
        }
        for dev in devs.keys():
            self.assertEqual(lvm2thin.pool_name_from_vg_lv(dev), devs[dev])

    def test_pool_name_from_vg_lv_bad_lv(self):
        lvm2thin = lvm2.Lvm2Thin(log)
        with self.assertRaises(SnapmCalloutError) as cm:
            lvm2thin.pool_name_from_vg_lv("some/lv")

    def test_pool_free_space(self):
        lvm2thin = lvm2.Lvm2Thin(log)
        pools = {
            ("fedora", "pool0"): 933940639,
            ("fedora", "pool1"): 1073741824,
            ("fedora", "pool2"): None,
        }
        for pool in pools.keys():
            if pools[pool] is not None:
                self.assertEqual(lvm2thin.pool_free_space(pool[0], pool[1]), pools[pool])
            else:
                with self.assertRaises(SnapmCalloutError) as cm:
                    lvm2thin.pool_free_space(pool[0], pool[1])

    def test_vg_free_space(self):
        lvm2cow = lvm2.Lvm2Cow(log)
        groups = {
            "fedora": 9097445376,
            "vg_hex": 16903045120,
            "appdata": 0,
            "nosuch": -1,
        }
        for vg in groups.keys():
            if groups[vg] != -1:
                self.assertEqual(lvm2cow.vg_free_space(vg), groups[vg])
            else:
                with self.assertRaises(SnapmCalloutError) as cm:
                    lvm2cow.vg_free_space(vg)

    def test_lvm2cow_discover_snapshots(self):
        lvm2cow = lvm2.Lvm2Cow(log)
        snapshots = lvm2cow.discover_snapshots()
        # FIXME: hardcoded value based on test data
        self.assertEqual(len(snapshots), 11)

    def test_lvm2thin_discover_snapshots(self):
        lvm2thin = lvm2.Lvm2Thin(log)
        snapshots = lvm2thin.discover_snapshots()
        # FIXME: hardcoded value based on test data
        self.assertEqual(len(snapshots), 5)



# Copyright Red Hat
#
# tests/test_snapm.py - snapm package unit tests
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
import unittest
import logging
from uuid import UUID

import snapm

from tests import MockArgs

log = logging.getLogger()

ONE_GIB = 2**30
TWO_GIB = 2 * 2**30
FOUR_GIB = 4 * 2**30
FIVE_GIB = 5 * 2**30
EIGHT_GIB = 8 * 2**30
TEN_GIB = 10 * 2**30

ONE_TIB = 2**40
TWO_TIB = 2 * 2**40
FOUR_TIB = 4 * 2**40
FIVE_TIB = 5 * 2**40
EIGHT_TIB = 8 * 2**40
TEN_TIB = 10 * 2**40


class SnapmTests(unittest.TestCase):
    """Test snapm module"""

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)

    def tearDown(self):
        log.debug("Tearing down (%s)", self._testMethodName)

    def test_set_debug_mask(self):
        snapm.set_debug_mask(snapm.SNAPM_DEBUG_ALL)

    def test_set_debug_mask_bad_mask(self):
        with self.assertRaises(ValueError) as cm:
            snapm.set_debug_mask(snapm.SNAPM_DEBUG_ALL + 1)

    def test_SnapmLogger(self):
        sl = snapm.SnapmLogger("snapm", 0)
        sl.debug("debug")

    def test_SnapmLogger_set_debug_mask(self):
        sl = snapm.SnapmLogger("snapm", 0)
        sl.set_debug_mask(snapm.SNAPM_DEBUG_ALL)

    def test_SnapmLogger_set_debug_mask_bad_mask(self):
        sl = snapm.SnapmLogger("snapm", 0)
        with self.assertRaises(ValueError) as cm:
            sl.set_debug_mask(snapm.SNAPM_DEBUG_ALL + 1)

    def test_SnapmLogger_debug_masked(self):
        sl = snapm.SnapmLogger("snapm", 0)
        snapm.set_debug_mask(snapm.SNAPM_DEBUG_ALL)
        sl.set_debug_mask(snapm.SNAPM_DEBUG_MANAGER)
        sl.debug_masked("quux")

    def test_Selection_is_null(self):
        s = snapm.Selection()
        self.assertTrue(s.is_null())

    def test_Selection_is_single_name(self):
        s = snapm.Selection(name="name")
        self.assertTrue(s.is_single())

    def test_Selection_is_single_uuid(self):
        s = snapm.Selection(uuid=UUID("2a6fd226-b400-577f-afe7-0c1a39c78488"))
        self.assertTrue(s.is_single())

    def test_Selection_is_not_single(self):
        s = snapm.Selection(origin="/dev/fedora/root")
        self.assertFalse(s.is_single())

    def test_Selection_is_not_null(self):
        s = snapm.Selection(name="name")
        self.assertFalse(s.is_null())

    def test_Selection_from_cmd_args_identifier_name(self):
        cmd_args = MockArgs()
        cmd_args.identifier = "name"
        s = snapm.Selection.from_cmd_args(cmd_args)
        self.assertEqual(cmd_args.identifier, s.name)

    def test_Selection_from_cmd_args_identifier_uuid(self):
        cmd_args = MockArgs()
        cmd_args.identifier = "2a6fd226-b400-577f-afe7-0c1a39c78488"
        s = snapm.Selection.from_cmd_args(cmd_args)
        self.assertEqual(UUID(cmd_args.identifier), s.uuid)

    def test_Selection_from_cmd_args_name(self):
        cmd_args = MockArgs()
        cmd_args.name = "name"
        s = snapm.Selection.from_cmd_args(cmd_args)
        self.assertEqual(cmd_args.name, s.name)

    def test_Selection_from_cmd_args_uuid(self):
        cmd_args = MockArgs()
        cmd_args.uuid = UUID("2a6fd226-b400-577f-afe7-0c1a39c78488")
        s = snapm.Selection.from_cmd_args(cmd_args)
        self.assertEqual(cmd_args.uuid, s.uuid)

    def test_valid_selection_snapset(self):
        s = snapm.Selection(name="testset0", timestamp=1693921253)
        s.check_valid_selection(snapshot_set=True)

    def test_valid_selection_snapshot(self):
        s = snapm.Selection(origin="/dev/fedora/home", mount_point="/home")
        s.check_valid_selection(snapshot=True)

    def test_valid_selection_snapset_invalid(self):
        s = snapm.Selection(origin="/dev/fedora/home")
        with self.assertRaises(ValueError) as cm:
            s.check_valid_selection(snapshot_set=True)

    def test_valid_selection_snapshot_invalid(self):
        s = snapm.Selection(name="testset0")
        with self.assertRaises(ValueError) as cm:
            s.check_valid_selection(snapshot=True)

    def test_valid_size_policy_fixed(self):
        policy = snapm.SizePolicy("/", "/", TEN_GIB, FOUR_GIB, TEN_GIB, "2G")
        self.assertEqual(policy.size, TWO_GIB)
        policy = snapm.SizePolicy("/", "/", TEN_TIB, FOUR_TIB, TEN_TIB, "2T")
        self.assertEqual(policy.size, TWO_TIB)
        policy = snapm.SizePolicy("/", "/", TEN_GIB, FOUR_GIB, TEN_GIB, "2GiB")
        self.assertEqual(policy.size, TWO_GIB)
        policy = snapm.SizePolicy("/", "/", TEN_TIB, FOUR_TIB, TEN_TIB, "2TiB")
        self.assertEqual(policy.size, TWO_TIB)

    def test_valid_size_policy_free(self):
        policy = snapm.SizePolicy("/", "/", TEN_GIB, FOUR_GIB, TEN_GIB, "10%FREE")
        self.assertEqual(policy.size, ONE_GIB)
        policy = snapm.SizePolicy("/", "/", TEN_GIB, FOUR_GIB, TEN_GIB, "50%FREE")
        self.assertEqual(policy.size, FIVE_GIB)
        policy = snapm.SizePolicy("/", "/", TEN_TIB, FOUR_TIB, TEN_TIB, "80%FREE")
        self.assertEqual(policy.size, EIGHT_TIB)
        policy = snapm.SizePolicy("/", "/", TEN_TIB, FOUR_TIB, TEN_TIB, "100%FREE")
        self.assertEqual(policy.size, TEN_TIB)

    def test_valid_size_policy_used(self):
        policy = snapm.SizePolicy("/", "/", TEN_GIB, FOUR_GIB, TEN_GIB, "25%USED")
        self.assertEqual(policy.size, ONE_GIB)
        policy = snapm.SizePolicy("/", "/", TEN_GIB, FOUR_GIB, TEN_GIB, "50%USED")
        self.assertEqual(policy.size, TWO_GIB)
        policy = snapm.SizePolicy("/", "/", TEN_TIB, FOUR_TIB, TEN_TIB, "100%USED")
        self.assertEqual(policy.size, FOUR_TIB)
        policy = snapm.SizePolicy("/", "/", TEN_TIB, FOUR_TIB, TEN_TIB, "200%USED")
        self.assertEqual(policy.size, EIGHT_TIB)

    def test_valid_size_policy_size(self):
        policy = snapm.SizePolicy("/", "/", TEN_GIB, FOUR_GIB, TEN_GIB, "20%SIZE")
        self.assertEqual(policy.size, TWO_GIB)
        policy = snapm.SizePolicy("/", "/", TEN_GIB, FOUR_GIB, TEN_GIB, "50%SIZE")
        self.assertEqual(policy.size, FIVE_GIB)
        policy = snapm.SizePolicy("/", "/", TEN_TIB, FOUR_TIB, TEN_TIB, "80%SIZE")
        self.assertEqual(policy.size, EIGHT_TIB)
        policy = snapm.SizePolicy("/", "/", TEN_TIB, FOUR_TIB, TEN_TIB, "100%SIZE")
        self.assertEqual(policy.size, TEN_TIB)

    def test_size_policy_size_over_limit_raises(self):
        with self.assertRaises(snapm.SnapmSizePolicyError) as cm:
            policy = snapm.SizePolicy("/", "/", TEN_TIB, FOUR_TIB, TEN_TIB, "200%SIZE")

    def test_size_policy_fixed_bad_unit_raises(self):
        with self.assertRaises(snapm.SnapmSizePolicyError) as cm:
            policy = snapm.SizePolicy("/", "/", TEN_GIB, FOUR_GIB, TEN_GIB, "2A")

    def test_size_policy_invalid_policy_raises(self):
        with self.assertRaises(snapm.SnapmSizePolicyError) as cm:
            policy = snapm.SizePolicy("/", "/", TEN_GIB, FOUR_GIB, TEN_GIB, "100%QUX")

    def test_is_size_policy_valid(self):
        self.assertEqual(True, snapm.is_size_policy("2G"))
        self.assertEqual(True, snapm.is_size_policy("100%FREE"))
        self.assertEqual(True, snapm.is_size_policy("50%USED"))
        self.assertEqual(True, snapm.is_size_policy("100%SIZE"))

    def test_is_size_policy_invalid(self):
        self.assertEqual(False, snapm.is_size_policy("2A"))
        self.assertEqual(False, snapm.is_size_policy("100%QUUX"))
        self.assertEqual(False, snapm.is_size_policy("foo"))
        self.assertEqual(False, snapm.is_size_policy("100%"))



# Copyright Red Hat
#
# tests/test_command.py - CLI layer tests
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
import unittest
import logging
import os

log = logging.getLogger()

import snapm
import snapm.command as command
from snapm.report import ReportOpts
import snapm.manager
import boom

from tests import MockArgs, have_root, BOOT_ROOT_TEST

from ._util import LvmLoopBacked, StratisLoopBacked


boom.set_boot_path(BOOT_ROOT_TEST)


class CommandTestsSimple(unittest.TestCase):
    """
    Test command interfaces
    """

    _old_path = None

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)
        bin_path = os.path.abspath("tests/bin")
        cur_path = os.environ["PATH"]
        self._old_path = cur_path
        os.environ["PATH"] = bin_path + os.pathsep + cur_path

    def tearDown(self):
        log.debug("Tearing down %s", self._testMethodName)
        os.environ["PATH"] = self._old_path

    def test__str_indent(self):
        test_str = "a\nb\nc\nd"
        xstr = "    a\n    b\n    c\n    d"
        output = command._str_indent(test_str, 4)
        self.assertEqual(output, xstr)

    def test__expand_fields_default(self):
        default_fields = "one,two,three"
        xfields = default_fields
        output = command._expand_fields(default_fields, "")
        self.assertEqual(output, xfields)

    def test__expand_fields_add(self):
        default_fields = "one,two,three"
        user_fields = "+four"
        xfields = default_fields + ",four"
        output = command._expand_fields(default_fields, user_fields)
        self.assertEqual(output, xfields)

    def test__expand_fields_replace(self):
        default_fields = "one,two,three"
        user_fields = "four,five,six"
        xfields = user_fields
        output = command._expand_fields(default_fields, user_fields)
        self.assertEqual(output, xfields)

    def test__report_opts_from_args_none(self):
        args = MockArgs()
        xopts = ReportOpts()
        opts = command._report_opts_from_args(args)
        self.assertEqual(opts, xopts)

    def test__report_opts_from_args_no_args(self):
        args = None
        xopts = ReportOpts()
        opts = command._report_opts_from_args(args)
        self.assertEqual(opts, xopts)

    def test__report_opts_from_args(self):
        args = MockArgs()
        args.rows = True
        args.separator = ":"
        args.name_prefixes = True
        args.no_headings = True

        xopts = ReportOpts()
        xopts.columns_as_rows = True
        xopts.separator = ":"
        xopts.field_name_prefix = "SNAPM_"
        xopts.unquoted = False
        xopts.aligned = False
        xopts.headings = False

        opts = command._report_opts_from_args(args)

        self.assertEqual(opts, xopts)

    def test__list_cmd_simple(self):
        args = MockArgs()
        command._list_cmd(args)

    def test__list_cmd_simple_bad_field(self):
        args = MockArgs()
        args.options = "nosuchfield"
        command._list_cmd(args)

    def test__list_cmd_simple_debug(self):
        args = MockArgs()
        args.debug = "all"
        command._list_cmd(args)

    def test__list_cmd_simple_debug_bad_field(self):
        args = MockArgs()
        args.debug = "all"
        args.options = "nosuchfield"
        with self.assertRaises(ValueError) as cm:
            command._list_cmd(args)

    def test__list_cmd_simple_verbose(self):
        args = MockArgs()
        args.verbose = 1
        command._list_cmd(args)

    def test__list_cmd_simple_fields(self):
        args = MockArgs()
        args.options = "name,uuid"
        command._list_cmd(args)

    def test__list_snapshot_cmd_simple(self):
        args = MockArgs()
        command._snapshot_list_cmd(args)

    def test__show_cmd_simple(self):
        args = MockArgs()
        command._show_cmd(args)

    def test__show_cmd_simple_verbose(self):
        args = MockArgs()
        args.verbose = 1
        command._show_cmd(args)

    def test__show_snapshot_cmd_simple(self):
        args = MockArgs()
        command._snapshot_show_cmd(args)

    def test__plugin_list_cmd_simple(self):
        args = MockArgs()
        command._plugin_list_cmd(args)

    def test_main_plugin_list(self):
        args = [os.path.join(os.getcwd(), "bin/snapm"), "plugin", "list"]
        self.assertEqual(command.main(args), 0)

    def test_set_debug_none(self):
        args = MockArgs()
        command.set_debug(args.debug)

    def test_set_debug_single(self):
        from snapm import get_debug_mask, SNAPM_DEBUG_COMMAND

        args = MockArgs()
        args.debug = "command"
        command.set_debug(args.debug)
        self.assertEqual(get_debug_mask(), SNAPM_DEBUG_COMMAND)

    def test_set_debug_list(self):
        from snapm import get_debug_mask, SNAPM_DEBUG_ALL

        args = MockArgs()
        args.debug = "manager,command,report"
        command.set_debug(args.debug)
        self.assertEqual(get_debug_mask(), SNAPM_DEBUG_ALL)

    def test_set_debug_all(self):
        from snapm import get_debug_mask, SNAPM_DEBUG_ALL

        args = MockArgs()
        args.debug = "all"
        command.set_debug(args.debug)
        self.assertEqual(get_debug_mask(), SNAPM_DEBUG_ALL)

    def test_set_debug_single_bad(self):
        args = MockArgs()
        args.debug = "nosuch"
        with self.assertRaises(ValueError) as cm:
            command.set_debug(args.debug)

    def test_main_version(self):
        args = [os.path.join(os.getcwd(), "bin/snapm"), "--version"]
        with self.assertRaises(SystemExit) as cm:
            command.main(args)

    def test_main_too_few_args(self):
        args = [os.path.join(os.getcwd(), "bin/snapm")]
        r = command.main(args)
        self.assertEqual(r, 1)

    def test_main_bad_command_type(self):
        args = [os.path.join(os.getcwd(), "bin/snapm"), "nosuch", "command"]
        with self.assertRaises(SystemExit) as cm:
            command.main(args)

    def test_main_bad_command(self):
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "nocommand"]
        with self.assertRaises(SystemExit) as cm:
            command.main(args)


@unittest.skipIf(not have_root(), "requires root privileges")
class CommandTests(unittest.TestCase):
    """
    Test command interfaces with devices
    """

    volumes = ["root", "home", "var"]
    thin_volumes = ["opt", "srv"]
    stratis_volumes = ["fs1", "fs2"]

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)
        def cleanup_lvm():
            log.debug("Cleaning up LVM (%s)", self._testMethodName)
            if hasattr(self, "_lvm"):
                self._lvm.destroy()

        def cleanup_stratis():
            log.debug("Cleaning up Stratis (%s)", self._testMethodName)
            if hasattr(self, "_stratis"):
                self._stratis.destroy()

        self.addCleanup(cleanup_lvm)
        self.addCleanup(cleanup_stratis)

        self._lvm = LvmLoopBacked(self.volumes, thin_volumes=self.thin_volumes)
        self._stratis = StratisLoopBacked(self.stratis_volumes)

        self.manager = snapm.manager.Manager()

    def mount_points(self):
        return self._lvm.mount_points() + self._stratis.mount_points()

    def test_print_snapsets(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        command.print_snapsets(self.manager)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_print_snapshots(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        command.print_snapshots(self.manager)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_show_snapsets(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        command.show_snapsets(self.manager)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_show_snapsets_with_members(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        command.show_snapsets(self.manager, members=True)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_show_snapsets_with_selection(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="testset0")
        command.show_snapsets(self.manager, selection=s)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_show_snapshots(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        command.show_snapshots(self.manager)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapset(self):
        command.create_snapset(self.manager, "testset0", self.mount_points())
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapset_default_size_policy(self):
        command.create_snapset(
            self.manager, "testset0", self.mount_points(), size_policy="10%FREE"
        )
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapset_per_mount_size_policy_10_free(self):
        mount_specs = [f"{mp}:10%FREE" for mp in self.mount_points()]
        command.create_snapset(self.manager, "testset0", mount_specs)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapset_per_mount_size_policy_100_size(self):
        mount_specs = [f"{mp}:100%SIZE" for mp in self.mount_points()]
        command.create_snapset(self.manager, "testset0", mount_specs)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_delete_snapset(self):
        command.create_snapset(self.manager, "testset0", self.mount_points())
        command.delete_snapset(self.manager, snapm.Selection(name="testset0"))

    def test_delete_snapset_ambiguous(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        with self.assertRaises(snapm.SnapmInvalidIdentifierError) as cm:
            command.delete_snapset(self.manager, snapm.Selection(nr_snapshots=2))
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_rename_snapset(self):
        command.create_snapset(self.manager, "testset0", self.mount_points())
        command.rename_snapset(self.manager, "testset0", "testset1")
        sets = self.manager.find_snapshot_sets(snapm.Selection(name="testset1"))
        self.assertEqual(len(sets), 1)
        self.assertEqual(sets[0].name, "testset1")
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset1"))

    def test_main_snapset_create(self):
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "create", "testset0"]
        args.extend(self.mount_points())
        command.main(args)

        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "delete", "testset0"]
        command.main(args)

    def test_main_snapset_delete(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "delete", "testset0"]
        command.main(args)

    def test_main_snapset_rename(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [
            os.path.join(os.getcwd(), "bin/snapm"),
            "snapset",
            "rename",
            "testset0",
            "testset1",
        ]
        command.main(args)

        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "delete", "testset1"]
        command.main(args)

    def test_main_snapset_rename_missing_newname(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "rename", "testset0"]
        with self.assertRaises(SystemExit) as cm:
            command.main(args)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapset_split(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        to_split = self.mount_points()[0]
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "split"]
        args.append("testset0")
        args.append("testset1")
        args.append(to_split)
        command.main(args)

        # Refresh manager context
        self.manager.discover_snapshot_sets()

        # Verify split
        sets = self.manager.find_snapshot_sets(snapm.Selection(name="testset1"))
        self.assertEqual(len(sets), 1)
        self.assertTrue(to_split in sets[0].sources)

    def test_main_snapset_prune(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        to_prune = self.mount_points()[0]
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "prune"]
        args.append("testset0")
        args.append(to_prune)
        command.main(args)

        # Refresh manager context
        self.manager.discover_snapshot_sets()

        # Verify prune
        pruned = self.manager.find_snapshot_sets(snapm.Selection(name="testset0"))[0]
        self.assertFalse(to_prune in pruned.sources)

    def test_main_snapset_list(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "list"]
        command.main(args)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapset_list_verbose(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "-v", "snapset", "list"]
        command.main(args)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapset_list_very_verbose_debug(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [
            os.path.join(os.getcwd(), "bin/snapm"),
            "-vv",
            "--debug=all",
            "snapset",
            "list",
        ]
        command.main(args)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapset_list_bad_debug(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [
            os.path.join(os.getcwd(), "bin/snapm"),
            "-vv",
            "--debug=quux",
            "snapset",
            "list",
        ]
        r = command.main(args)
        self.assertEqual(r, 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapshot_activate(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapshot", "activate"]
        self.assertEqual(command.main(args), 0)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapshot_deactivate(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapshot", "deactivate"]
        self.assertEqual(command.main(args), 0)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapshot_autoactivate(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [
            os.path.join(os.getcwd(), "bin/snapm"),
            "snapshot",
            "autoactivate",
            "--yes",
        ]
        self.assertEqual(command.main(args), 0)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapshot_activate_nosuch(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [
            os.path.join(os.getcwd(), "bin/snapm"),
            "snapshot",
            "activate",
            "-N",
            "nosuch",
        ]
        self.assertEqual(command.main(args), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapshot_deactivate_nosuch(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [
            os.path.join(os.getcwd(), "bin/snapm"),
            "snapshot",
            "deactivate",
            "-N",
            "nosuch",
        ]
        self.assertEqual(command.main(args), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapshot_autoactivate_nosuch(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [
            os.path.join(os.getcwd(), "bin/snapm"),
            "snapshot",
            "autoactivate",
            "--yes",
            "-N",
            "nosuch",
        ]
        self.assertEqual(command.main(args), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapset_show(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "show"]
        self.assertEqual(command.main(args), 0)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapshot_list(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapshot", "list"]
        self.assertEqual(command.main(args), 0)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapshot_show(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapshot", "show"]
        self.assertEqual(command.main(args), 0)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapset_activate(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "activate"]
        self.assertEqual(command.main(args), 0)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapset_deactivate(self):
        self.manager.create_snapshot_set("testset0", self._lvm.mount_points())
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "deactivate"]
        self.assertEqual(command.main(args), 0)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapset_autoactivate(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        args = [
            os.path.join(os.getcwd(), "bin/snapm"),
            "snapset",
            "autoactivate",
            "--yes",
        ]
        self.assertEqual(command.main(args), 0)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_main_snapset_revert(self):
        # Revert is only supported on LVM2: don't include Stratis in the test.
        origin_file = "root/origin"
        snapshot_file = "root/snapshot"
        testset = "testset0"

        # Create files in the origin volume and post-snapshot
        self._lvm.touch_path(origin_file)
        self.manager.create_snapshot_set(testset, self.mount_points())
        self._lvm.touch_path(snapshot_file)

        # Test that the origin and snapshot files both exist
        self.assertEqual(self._lvm.test_path(origin_file), True)
        self.assertEqual(self._lvm.test_path(snapshot_file), True)

        # Start revert the snapshot set
        args = [os.path.join(os.getcwd(), "bin/snapm"), "snapset", "revert", testset]
        self.assertEqual(command.main(args), 0)

        # Unmount the set, deactivate/reactivate and re-mount
        self._lvm.umount_all()
        self._lvm.deactivate()
        self._lvm.activate()
        self._lvm.mount_all()

        self._stratis.umount_all()
        self._stratis.stop_pool()
        self._stratis.start_pool()
        self._stratis.mount_all()

        # Test that only the origin file exists
        self.assertEqual(self._lvm.test_path(origin_file), True)
        self.assertEqual(self._lvm.test_path(snapshot_file), False)



# Copyright Red Hat
#
# snapm/manager/boot.py - Snapshot Manager boot support
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
"""
Boot integration for snapshot manager
"""
from os import uname
from os.path import exists as path_exists
import logging

import boom
import boom.cache
import boom.command
from boom.config import load_boom_config
from boom.bootloader import (
    OPTIONAL_KEYS,
    optional_key_default,
    key_to_bls_name,
)
from boom.osprofile import match_os_profile_by_version

from snapm import (
    SnapmNotFoundError,
    SnapmCalloutError,
    ETC_FSTAB,
)

_log = logging.getLogger(__name__)

_log_debug = _log.debug
_log_info = _log.info
_log_warn = _log.warning
_log_error = _log.error

#: Path to the system machine-id file
_MACHINE_ID = "/etc/machine-id"
#: Path to the legacy system machine-id file
_DBUS_MACHINE_ID = "/var/lib/dbus/machine-id"
#: Path to the system os-release file
_OS_RELEASE = "/etc/os-release"

#: Snapshot set kernel command line argument
SNAPSET_ARG = "snapm.snapset"

#: Snapshot set revert kernel command line argument
REVERT_ARG = "snapm.revert"

#: /dev path prefix
_DEV_PREFIX = "/dev/"


def _get_uts_release():
    """
    Return the UTS release (kernel version) of the running system.

    :returns: A string representation of the UTS release value.
    """
    return uname()[2]


def _get_machine_id():
    """
    Return the current host's machine-id.

    Get the machine-id value for the running system by reading from
    ``/etc/machine-id`` and return it as a string.

    :returns: The ``machine_id`` as a string
    :rtype: str
    """
    if path_exists(_MACHINE_ID):
        path = _MACHINE_ID
    elif path_exists(_DBUS_MACHINE_ID):
        path = _DBUS_MACHINE_ID
    else:
        return None

    with open(path, "r", encoding="utf8") as file:
        try:
            machine_id = file.read().strip()
        except OSError as err:
            _log_error("Could not read machine-id from '%s': %s", path, err)
            machine_id = None
    return machine_id


def _find_snapset_root(snapset):
    """
    Find the device that backs the root filesystem for snapshot set ``snapset``.

    If the snapset does not include the root volume look the device up via the
    fstab.
    """
    for snapshot in snapset.snapshots:
        if snapshot.mount_point == "/":
            return snapshot
    # Note: add fstab lookup for non-root snapsets
    # needs either root=UUID/LABEL support in boom or a lookup to resolve any
    # UUID/LABEL found in the file.
    raise SnapmNotFoundError(f"Could not find root device for snapset {snapset.name}")


def _create_default_os_profile():
    """
    Create a default boom OsProfile for the running system.
    This uses the boom API to run the equivalent of:
      `boom profile create --from-host`.
    """
    options = boom.command.os_options_from_cmdline()
    _log_info(
        "Creating default boot profile from %s with options %s", _OS_RELEASE, options
    )
    return boom.command.create_profile(
        None, None, None, None, profile_file=_OS_RELEASE, options=options
    )


def _build_snapset_mount_list(snapset):
    """
    Build a list of command line mount unit definitions for the snapshot set
    ``snapset``. Mount points that are not part of the snapset are substituted
    from /etc/fstab.

    :param snapset: The snapshot set to build a mount list for.
    """
    mounts = []
    snapset_mounts = snapset.mount_points
    with open(ETC_FSTAB, "r", encoding="utf8") as fstab:
        for line in fstab.readlines():
            if line == "\n" or line.startswith("#"):
                continue
            what, where, fstype, options, _, _ = line.split()
            if where == "/":
                continue
            if where in snapset_mounts:
                snapshot = snapset.snapshot_by_mount_point(where)
                mounts.append(f"{snapshot.devpath}:{where}:{fstype}:{options}")
            else:
                mounts.append(f"{what}:{where}:{fstype}:{options}")
    return mounts


def _create_boom_boot_entry(
    version, title, tag_arg, root_device, lvm_root_lv=None, mounts=None
):
    """
    Create a boom boot entry according to the passed arguments.

    :param version: The UTS release name for the boot entry
    :param title: The title for the boot entry.
    :param tag_arg: A tag argument to be added to the kernel command line and
                    used to associate the entry with a snapshot set name or
                    UUID.
    :param root_device: The root device for the entry. Passed to root=...
    :param lvm_root_lv: An optional LVM2 root logical volume.
    :param mounts: An optional list of mount specifications to use for the
                   boot entry. If defined fstab=no will be appended to the
                   generated kernel command line.
    """
    assert title is not None, "Boot entry argument title must have a value"
    assert version is not None, "Boot entry argument version must have a value"
    assert tag_arg is not None, "Boot entry argument tag_arg must have a value"
    assert root_device is not None, "Boot entry argument root_device must have a value"

    machine_id = _get_machine_id()

    osp = match_os_profile_by_version(version)
    if not osp:
        try:
            osp = _create_default_os_profile()
        except ValueError as err:
            raise SnapmCalloutError(
                f"Error calling boom to create default OsProfile: {err}"
            ) from err

    if mounts:
        add_opts = f"rw {tag_arg}"
        del_opts = "ro"
    else:
        add_opts = tag_arg
        del_opts = None

    entry = boom.command.create_entry(
        title,
        version,
        machine_id,
        root_device,
        lvm_root_lv=lvm_root_lv,
        profile=osp,
        add_opts=add_opts,
        del_opts=del_opts,
        write=False,
        images=boom.command.I_BACKUP,
        no_fstab=bool(mounts),
        mounts=mounts,
    )

    # Apply defaults for optional keys enabled in profile
    for opt_key in OPTIONAL_KEYS:
        bls_key = key_to_bls_name(opt_key)
        if bls_key in osp.optional_keys:
            setattr(entry, bls_key, optional_key_default(opt_key))

    # Write BLS snippet for entry
    entry.write_entry()

    return entry


def create_snapset_boot_entry(snapset, title=None):
    """
    Create a boom boot entry to boot into the snapshot set represented by
    ``snapset``.

    :param snapset: The snapshot set for which to create a boot entry.
    :param title: An optional title for the boot entry. If ``title`` is
                  ``None`` the boot entry will be titled as
                  "Snapshot snapset_name snapset_time".
    """
    version = _get_uts_release()
    title = title or f"Snapshot {snapset.name} {snapset.time} ({version})"
    root_snapshot = _find_snapset_root(snapset)
    root_device = root_snapshot.devpath
    if root_snapshot.provider.name in ("lvm2-cow", "lvm2-thin"):
        lvm_root_lv = root_snapshot.name
    else:
        lvm_root_lv = None
    mounts = _build_snapset_mount_list(snapset)
    snapset.boot_entry = _create_boom_boot_entry(
        version,
        title,
        f"{SNAPSET_ARG}={snapset.uuid}",
        root_device,
        lvm_root_lv=lvm_root_lv,
        mounts=mounts,
    )
    _log_debug(
        "Created boot entry '%s' for snapshot set with UUID=%s", title, snapset.uuid
    )


def create_snapset_revert_entry(snapset, title=None):
    """
    Create a boom boot entry to revert the snapshot set represented by
    ``snapset``.

    :param snapset: The snapshot set for which to create a revert entry.
    :param title: An optional title for the revert entry. If ``title`` is
                  ``None`` the revert entry will be titled as
                  "Revert snapset_name snapset_time".
    """
    version = _get_uts_release()
    title = title or f"Revert {snapset.name} {snapset.time} ({version})"
    root_snapshot = _find_snapset_root(snapset)
    root_device = root_snapshot.origin
    if root_snapshot.provider.name in ("lvm2-cow", "lvm2-thin"):
        lvm_root_lv = root_device.removeprefix(_DEV_PREFIX)
    else:
        lvm_root_lv = None
    snapset.revert_entry = _create_boom_boot_entry(
        version,
        title,
        f"{REVERT_ARG}={snapset.uuid}",
        root_device,
        lvm_root_lv=lvm_root_lv,
    )
    _log_debug(
        "Created revert entry '%s' for snapshot set with UUID=%s", title, snapset.uuid
    )


def _delete_boot_entry(boot_id):
    """
    Delete a boom boot entry by ID.

    :param boot_id: The boot identifier to delete.
    """
    selection = boom.Selection(boot_id=boot_id)
    boom.command.delete_entries(selection=selection)
    boom.cache.clean_cache()


def delete_snapset_boot_entry(snapset):
    """
    Delete the boot entry corresponding to ``snapset``.

    :param snapset: The snapshot set for which to remove a boot entry.
    """
    if snapset.boot_entry is None:
        return
    _delete_boot_entry(snapset.boot_entry.boot_id)


def delete_snapset_revert_entry(snapset):
    """
    Delete the revert entry corresponding to ``snapset``.

    :param snapset: The snapshot set for which to remove a revert entry.
    """
    if snapset.revert_entry is None:
        return
    _delete_boot_entry(boot_id=snapset.revert_entry.boot_id)


def check_boom_config():
    """
    Check for boom configuration and create the default config if not found.
    """
    try:
        load_boom_config()
    except ValueError:
        _log_info("No boom configuration found: generating default config in /boot")
        boom.command.create_config()


class BootEntryCache(dict):
    """
    Cache mapping snapshot sets to boom ``BootEntry`` instances.

    Boot entries in the cache are either snapshot set boot entries or revert
    entries depending on the value of ``entry_arg``.
    """

    def __init__(self, entry_arg):
        super().__init__()
        self.entry_arg = entry_arg
        self.refresh_cache()

    def _parse_entry(self, boot_entry):
        """
        Parse a boom ``BootEntry`` options string and return the value
        of the ``snapm.snapset`` argument if present.

        :param boot_entry: The boot entry to process.
        :returns: The snapset name for the boot entry.
        """
        for word in boot_entry.options.split():
            if word.startswith(self.entry_arg):
                _, value = word.split("=")
                return value
        return None

    def refresh_cache(self):
        """
        Generate mappings from snapshot sets to boot entries.
        """
        self.clear()
        entries = boom.command.find_entries()
        for boot_entry in entries:
            snapset = self._parse_entry(boot_entry)
            if snapset:
                self[snapset] = boot_entry


class BootCache:
    """
    Set of caches mapping snapshot sets to boot entries and revert entries.
    """

    def __init__(self):
        self.entry_cache = BootEntryCache(SNAPSET_ARG)
        _log_debug(
            "Initialised boot entry cache with %d entries", len(self.entry_cache)
        )
        self.revert_cache = BootEntryCache(REVERT_ARG)
        _log_debug(
            "Initialised revert boot entry cache with %d entries",
            len(self.revert_cache),
        )

    def refresh_cache(self):
        """
        Refresh the cache of boot entry mappings held by this ``BootCache``
        instance.
        """
        self.entry_cache.refresh_cache()
        _log_debug("Refreshed boot entry cache with %d entries", len(self.entry_cache))
        self.revert_cache.refresh_cache()
        _log_debug(
            "Refreshed revert boot entry cache with %d entries",
            len(self.revert_cache),
        )



# Copyright Red Hat
#
# tests/test_stratis.py - Stratis plugin unit tests
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
import unittest
import logging
import os.path
import psutil
import os
from subprocess import run

from dbus_client_gen import DbusClientUniqueResultError

log = logging.getLogger()

from snapm.manager.plugins.stratislib import (
    get_object,
    TOP_OBJECT,
    ObjectManager,
)
from snapm import SnapmNotFoundError
import snapm.manager.plugins.stratis as stratis
from snapm.manager.plugins import device_from_mount_point

from tests import have_root
from ._util import StratisLoopBacked


def _systemctl_args(command):
    return [
        "systemctl",
        f"{command}",
        "stratisd"
    ]


class StratisTestsSimple(unittest.TestCase):
    """Test Stratis plugin functions"""

    _old_path = None

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)

        def cleanup():
            log.debug("Cleaning up (%s)", self._testMethodName)
            os.environ["PATH"] = self._old_path

        self.addCleanup(cleanup)

        bin_path = os.path.abspath("tests/bin")
        cur_path = os.environ["PATH"]
        self._old_path = cur_path
        os.environ["PATH"] = bin_path + os.pathsep + cur_path

    def test_is_stratis_device(self):
        devs = {
            "/dev/mapper/fedora-home": False,
            "/dev/mapper/fedora-root": False,
            "/dev/mapper/fedora-var": False,
            "/dev/mapper/stratis-1-1c7c941a2dba4eb78d57d3fb01aacc61-thin-fs-202ea1667fa54123bd24f0c353b9914c": True,
            "/dev/stratis/p1/fs1": True,
            "/dev/mapper/mpatha": False,
            "/dev/vda2": False,
            "/dev/quux": False,
        }
        for dev in devs.keys():
            self.assertEqual(stratis.is_stratis_device(dev), devs[dev])

    def pool_fs_from_origin(self):
        devs = {
            "/dev/stratis/p1/fs1": ("p1", "fs1"),
            "/dev/stratis/p1/fs2": ("p1", "fs2"),
            "/dev/stratis/pool0/fs0": ("pool0", "fs0"),
            "/dev/stratis/data/vol0": ("data", "vol0"),
        }
        for dev in devs.keys():
            self.assertEqual(stratis.pool_fs_from_origin(dev), devs[dev])

    def test_is_stratisd_running(self):
        running = False
        for proc in psutil.process_iter():
            try:
                if proc.name() == "stratisd":
                    running = True
            except psutil.NoSuchProcess:
                running = False
        self.assertEqual(running, stratis.is_stratisd_running())

    @unittest.skipIf(not have_root(), "requires root privileges")
    def test_is_stratisd_running_stopped(self):
        systemctl_stop_args = _systemctl_args("stop")
        run(systemctl_stop_args, check=True)
        self.assertEqual(False, stratis.is_stratisd_running())
        systemctl_start_args = _systemctl_args("start")
        run(systemctl_start_args, check=True)

    def test_stratis_devices_present(self):
        self.assertEqual(stratis.stratis_devices_present(), True)

    def test__snapshot_min_size(self):
        sizes = [
            (256 * 2**20, 512 * 2**20),  # 256MiB/512MiB
            (512 * 2**20, 512 * 2**20),  # 512MiB/512MiB
            (1 * 2**30, 1 * 2**30),      # 1GiB/1GiB
            (100 * 2**30, 100 * 2**30),  # 100GiB/100GiB
            (1 * 2**40, 1 * 2**40),      # 1TiB/1TiB
        ]
        for (size, xsize) in sizes:
            self.assertEqual(xsize, stratis._snapshot_min_size(size))


@unittest.skipIf(not have_root(), "requires root privileges")
class StratisTests(unittest.TestCase):

    stratis_volumes = ["fs1", "fs2"]

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)
        def cleanup():
            log.debug("Cleaning up Stratis (%s)", self._testMethodName)
            if hasattr(self, "_stratis"):
                self._stratis.destroy()

        self.addCleanup(cleanup)

        self._stratis = StratisLoopBacked(self.stratis_volumes)

    def test__get_pool_filesystem(self):
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        (pool, filesystem) = stratis._get_pool_filesystem(managed_objects, "pool1", "fs1")
        self.assertTrue(str(pool.Name()) == "pool1")
        self.assertTrue(str(filesystem.Name()) == "fs1")

    def test__get_pool_filesystem_bad_pool(self):
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        with self.assertRaises(DbusClientUniqueResultError) as cm:
            (pool, filesystem) = stratis._get_pool_filesystem(managed_objects, "nosuchpool1", "fs1")

    def test__get_pool_filesystem_bad_fs(self):
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        with self.assertRaises(DbusClientUniqueResultError) as cm:
            (pool, filesystem) = stratis._get_pool_filesystem(managed_objects, "pool1", "nosuchfs1")

    def test_pool_free_space_bytes(self):
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        (pool, filesystem) = stratis._get_pool_filesystem(managed_objects, "pool1", "fs1")
        free_bytes = stratis._pool_free_space_bytes(managed_objects, "pool1")
        self.assertEqual(
            int(pool.TotalPhysicalSize()) - int(pool.TotalPhysicalUsed()[1]) if pool.TotalPhysicalUsed()[0] else 0,
            free_bytes
        )

    def test_fs_size_bytes(self):
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        size_bytes = stratis._fs_size_bytes(managed_objects, "pool1", "fs1")
        self.assertEqual(size_bytes, 2**30)

    def test_pool_fs_from_device_path(self):
        devpath = device_from_mount_point(self._stratis.mount_points()[0])
        (pool, fs) = stratis.pool_fs_from_device_path(devpath)
        self.assertEqual(pool, "pool1")
        self.assertEqual(fs, "fs1")

    def test_stratis_discover_snapshots(self):
        self._stratis.create_snapshot("fs1", "fs1-snapset_test_1721136677_-opt")
        self._stratis.create_snapshot("fs2", "fs2-snapset_test_1721136677_-data")
        stratis_plugin = stratis.Stratis(log)
        snapshots = stratis_plugin.discover_snapshots()
        self.assertEqual(len(snapshots), 2)

    def test_stratis_can_snapshot_no_stratisd(self):
        systemctl_stop_args = _systemctl_args("stop")
        run(systemctl_stop_args, check=True)

        with self.assertRaises(SnapmNotFoundError) as cm:
            stratis_plugin = stratis.Stratis(log)

        systemctl_start_args = _systemctl_args("start")
        run(systemctl_start_args, check=True)

    def test__origin_uuid_to_fs_name(self):
        self._stratis.create_snapshot("fs1", "fs1-snapset_test_1721136677_-opt")

        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        (pool, fs) = stratis._get_pool_filesystem(managed_objects, "pool1", "fs1-snapset_test_1721136677_-opt")

        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        origin = stratis._origin_uuid_to_fs_name(
            managed_objects, fs.Pool(), str(fs.Origin()[1])
        )

        self.assertEqual(origin, "fs1")

    def test_filter_stratis_snapshot_snapshot(self):
        self._stratis.create_snapshot("fs1", "fs1-snapset_test_1721136677_-opt")

        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        (pool, fs) = stratis._get_pool_filesystem(managed_objects, "pool1", "fs1-snapset_test_1721136677_-opt")
        self.assertEqual(True, stratis.filter_stratis_snapshot(fs))

    def test_filter_stratis_snapshot_nonsnapshot(self):
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        (pool, fs) = stratis._get_pool_filesystem(managed_objects, "pool1", "fs1")
        self.assertEqual(False, stratis.filter_stratis_snapshot(fs))
